@@ -1,3 +1,4 @@
+use crate::active_speaker_observer::ActiveSpeakerObserverOptions;
 use crate::audio_level_observer::AudioLevelObserverOptions;
 use crate::consumer::{
     ConsumerDump, ConsumerId, ConsumerLayers, ConsumerScore, ConsumerStats, ConsumerTraceEventType,
@@ -27,8 +28,8 @@ use crate::worker::{WorkerDump, WorkerUpdateSettings};
 use parking_lot::Mutex;
 use serde::de::DeserializeOwned;
 use serde::{Deserialize, Serialize};
+use serde_json::Value;
 use std::fmt::Debug;
-use std::marker::PhantomData;
 use std::net::IpAddr;
 use std::num::NonZeroU16;
 
@@ -148,32 +149,6 @@ macro_rules! request_response {
     };
 }
 
-macro_rules! request_response_generic {
-    (
-        $method: literal,
-        $request_struct_name: ident { $( $request_field_name: ident: $request_field_type: ty$(,)? )* },
-        $generic_response: ident,
-    ) => {
-        #[derive(Debug, Serialize)]
-        pub(crate) struct $request_struct_name<$generic_response>
-        where
-            $generic_response: Debug + DeserializeOwned,
-        {
-            $( pub(crate) $request_field_name: $request_field_type, )*
-            #[serde(skip)]
-            pub(crate) phantom_data: PhantomData<$generic_response>,
-        }
-
-        impl<$generic_response: Debug + DeserializeOwned> Request for $request_struct_name<$generic_response> {
-            type Response = $generic_response;
-
-            fn as_method(&self) -> &'static str {
-                $method
-            }
-        }
-    };
-}
-
 request_response!("worker.close", WorkerCloseRequest {});
 
 request_response!("worker.dump", WorkerDumpRequest {}, WorkerDump);
@@ -332,8 +307,9 @@ request_response!(
         data: RouterCreatePlainTransportData,
     },
     PlainTransportData {
-        rtcp_mux: bool,
-        comedia: bool,
+        // The following fields are present, but unused
+        // rtcp_mux: bool,
+        // comedia: bool,
         tuple: Mutex<TransportTuple>,
         rtcp_tuple: Mutex<Option<TransportTuple>>,
         sctp_parameters: Option<SctpParameters>,
@@ -414,6 +390,30 @@ request_response!(
     },
 );
 
+#[derive(Debug, Serialize)]
+#[serde(rename_all = "camelCase")]
+pub(crate) struct RouterCreateActiveSpeakerObserverData {
+    pub(crate) interval: u16,
+}
+
+impl RouterCreateActiveSpeakerObserverData {
+    pub(crate) fn from_options(
+        active_speaker_observer_options: &ActiveSpeakerObserverOptions,
+    ) -> Self {
+        Self {
+            interval: active_speaker_observer_options.interval,
+        }
+    }
+}
+
+request_response!(
+    "router.createActiveSpeakerObserver",
+    RouterCreateActiveSpeakerObserverRequest {
+        internal: RtpObserverInternal,
+        data: RouterCreateActiveSpeakerObserverData,
+    },
+);
+
 request_response!(
     "transport.close",
     TransportCloseRequest {
@@ -421,20 +421,20 @@ request_response!(
     },
 );
 
-request_response_generic!(
+request_response!(
     "transport.dump",
     TransportDumpRequest {
         internal: TransportInternal,
     },
-    Dump,
+    Value,
 );
 
-request_response_generic!(
+request_response!(
     "transport.getStats",
     TransportGetStatsRequest {
         internal: TransportInternal,
     },
-    Stats,
+    Value,
 );
 
 #[derive(Debug, Serialize)]
@@ -445,7 +445,7 @@ pub(crate) struct TransportConnectRequestWebRtcData {
 
 request_response!(
     "transport.connect",
-    TransportConnectRequestWebRtc {
+    TransportConnectWebRtcRequest {
         internal: TransportInternal,
         data: TransportConnectRequestWebRtcData,
     },
@@ -465,7 +465,7 @@ pub(crate) struct TransportConnectRequestPipeData {
 
 request_response!(
     "transport.connect",
-    TransportConnectRequestPipe {
+    TransportConnectPipeRequest {
         internal: TransportInternal,
         data: TransportConnectRequestPipeData,
     },
@@ -489,7 +489,7 @@ pub(crate) struct TransportConnectRequestPlainData {
 
 request_response!(
     "transport.connect",
-    TransportConnectRequestPlain {
+    TransportConnectPlainRequest {
         internal: TransportInternal,
         data: TransportConnectRequestPlainData,
     },
