@@ -23,13 +23,14 @@ use crate::{ortc, uuid_based_wrapper_type};
 use async_executor::Executor;
 use async_trait::async_trait;
 use event_listener_primitives::HandlerId;
+use hash_hasher::HashedMap;
 use log::{error, warn};
+use nohash_hasher::IntMap;
 use parking_lot::Mutex;
 use serde::de::DeserializeOwned;
 use serde::{Deserialize, Serialize};
-use std::collections::HashMap;
+use serde_json::Value;
 use std::fmt::Debug;
-use std::marker::PhantomData;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::Arc;
 use thiserror::Error;
@@ -40,6 +41,27 @@ uuid_based_wrapper_type!(
     TransportId
 );
 
+// We don't want this to be a public API, but have to use it like this to be able to still use as
+// trait object
+mod private {
+    use super::Transport;
+
+    #[doc(hidden)]
+    pub trait CloneTransport {
+        #[doc(hidden)]
+        fn clone_box(&self) -> Box<dyn Transport>;
+    }
+
+    impl<T> CloneTransport for T
+    where
+        T: Transport + Clone + 'static,
+    {
+        fn clone_box(&self) -> Box<dyn Transport> {
+            Box::new(self.clone())
+        }
+    }
+}
+
 /// Data contained in transport trace events.
 ///
 /// See also "trace" event in the [Debugging](https://mediasoup.org/documentation/v3/mediasoup/debugging#trace-Event)
@@ -82,11 +104,11 @@ pub enum TransportTraceEventType {
 #[doc(hidden)]
 pub struct RtpListener {
     /// Map from Ssrc (as string) to producer ID
-    pub mid_table: HashMap<String, ProducerId>,
+    pub mid_table: HashedMap<String, ProducerId>,
     /// Map from Ssrc (as string) to producer ID
-    pub rid_table: HashMap<String, ProducerId>,
+    pub rid_table: HashedMap<String, ProducerId>,
     /// Map from Ssrc (as string) to producer ID
-    pub ssrc_table: HashMap<String, ProducerId>,
+    pub ssrc_table: HashedMap<String, ProducerId>,
 }
 
 #[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Hash, Deserialize, Serialize)]
@@ -105,7 +127,7 @@ pub struct RecvRtpHeaderExtensions {
 #[doc(hidden)]
 pub struct SctpListener {
     /// Map from stream ID (as string) to data producer ID
-    stream_id_table: HashMap<String, DataProducerId>,
+    stream_id_table: HashedMap<String, DataProducerId>,
 }
 
 #[derive(Debug, Copy, Clone, PartialEq)]
@@ -128,7 +150,7 @@ pub(super) enum TransportType {
 ///
 /// For additional methods see [`TransportGeneric`].
 #[async_trait(?Send)]
-pub trait Transport: Debug + Send + Sync + CloneTransport {
+pub trait Transport: Debug + Send + Sync + private::CloneTransport {
     /// Transport id.
     #[must_use]
     fn id(&self) -> TransportId;
@@ -215,31 +237,31 @@ pub trait Transport: Debug + Send + Sync + CloneTransport {
     /// Callback is called when a new producer is created.
     fn on_new_producer(
         &self,
-        callback: Box<dyn Fn(&Producer) + Send + Sync + 'static>,
+        callback: Arc<dyn Fn(&Producer) + Send + Sync + 'static>,
     ) -> HandlerId;
 
     /// Callback is called when a new consumer is created.
     fn on_new_consumer(
         &self,
-        callback: Box<dyn Fn(&Consumer) + Send + Sync + 'static>,
+        callback: Arc<dyn Fn(&Consumer) + Send + Sync + 'static>,
     ) -> HandlerId;
 
     /// Callback is called when a new data producer is created.
     fn on_new_data_producer(
         &self,
-        callback: Box<dyn Fn(&DataProducer) + Send + Sync + 'static>,
+        callback: Arc<dyn Fn(&DataProducer) + Send + Sync + 'static>,
     ) -> HandlerId;
 
     /// Callback is called when a new data consumer is created.
     fn on_new_data_consumer(
         &self,
-        callback: Box<dyn Fn(&DataConsumer) + Send + Sync + 'static>,
+        callback: Arc<dyn Fn(&DataConsumer) + Send + Sync + 'static>,
     ) -> HandlerId;
 
     /// See [`Transport::enable_trace_event()`]
     fn on_trace(
         &self,
-        callback: Box<dyn Fn(&TransportTraceEventData) + Send + Sync + 'static>,
+        callback: Arc<dyn Fn(&TransportTraceEventData) + Send + Sync + 'static>,
     ) -> HandlerId;
 
     /// Callback is called when the router this transport belongs to is closed for whatever reason.
@@ -253,28 +275,9 @@ pub trait Transport: Debug + Send + Sync + CloneTransport {
     fn on_close(&self, callback: Box<dyn FnOnce() + Send + 'static>) -> HandlerId;
 }
 
-// We don't want this to be a public API, but have to use it like this to be able to still use as
-// trait object
-/// This is a private method, don't use it outside of the library
-#[doc(hidden)]
-pub trait CloneTransport {
-    /// This is a private method, don't use it outside of the library
-    #[doc(hidden)]
-    fn clone_transport(&self) -> Box<dyn Transport>;
-}
-
-impl<T> CloneTransport for T
-where
-    T: Transport + Clone + 'static,
-{
-    fn clone_transport(&self) -> Box<dyn Transport> {
-        Box::new(self.clone())
-    }
-}
-
 impl Clone for Box<dyn Transport> {
     fn clone(&self) -> Self {
-        self.clone_transport()
+        self.clone_box()
     }
 }
 
@@ -369,7 +372,7 @@ pub(super) trait TransportImpl: TransportGeneric {
 
     fn next_mid_for_consumers(&self) -> &AtomicUsize;
 
-    fn used_sctp_stream_ids(&self) -> &Mutex<HashMap<u16, bool>>;
+    fn used_sctp_stream_ids(&self) -> &Mutex<IntMap<u16, bool>>;
 
     fn cname_for_producers(&self) -> &Mutex<Option<String>>;
 
@@ -393,26 +396,24 @@ pub(super) trait TransportImpl: TransportGeneric {
         }
     }
 
-    async fn dump_impl(&self) -> Result<Self::Dump, RequestError> {
+    async fn dump_impl(&self) -> Result<Value, RequestError> {
         self.channel()
             .request(TransportDumpRequest {
                 internal: TransportInternal {
                     router_id: self.router().id(),
                     transport_id: self.id(),
                 },
-                phantom_data: PhantomData {},
             })
             .await
     }
 
-    async fn get_stats_impl(&self) -> Result<Vec<Self::Stat>, RequestError> {
+    async fn get_stats_impl(&self) -> Result<Value, RequestError> {
         self.channel()
             .request(TransportGetStatsRequest {
                 internal: TransportInternal {
                     router_id: self.router().id(),
                     transport_id: self.id(),
                 },
-                phantom_data: PhantomData {},
             })
             .await
     }
@@ -645,8 +646,7 @@ pub(super) trait TransportImpl: TransportGeneric {
 
         Ok(Consumer::new(
             consumer_id,
-            producer.id(),
-            producer.kind(),
+            producer,
             r#type,
             rtp_parameters,
             response.paused,
@@ -813,7 +813,7 @@ pub(super) trait TransportImpl: TransportGeneric {
             response.sctp_stream_parameters,
             response.label,
             response.protocol,
-            data_producer.id(),
+            data_producer,
             Arc::clone(self.executor()),
             self.channel().clone(),
             self.payload_channel().clone(),
